# by default, test our biggest configurations
ENGINE=j2c
OPT_FLAG=
MODEL=MostlyCopyingSplitRegions-B_M_F_H
THREADS=RealtimeJVM
IO=SelectSockets_PollingOther
TRANS=false

ifeq ($(OVM_ARCH), i686)
ARCH_JIT=simplejit
else
ARCH_JIT=NO_JIT
endif

# On some systems, we may need to pass the bootimage file name as an
# argument to ovm.  make IMG=img will at least allow regressions tests
# to work on such systems
IMG=

# make regression-test produces minimal results to stdout that can
# easily be processed by our test-runner.
ifeq ($(REGRESS),true)
SHOW=@echo
REDIR=> ../$@ 2>&1
REDIRA=>$(REDIR)

# remove build directory for passing tests
all: check clean-1

else
SHOW=@true
REDIR=
REDIRA=

all: check
endif

# Each invocation of make will build and test a particular
# configuration, storing the results in a file derived from the config
# parameters.  Multiple configurations can be tested by changing these
# 5 variables, as is done in regression-test
check: $(ENGINE)$(OPT_FLAG)-$(MODEL)-$(IO)-$(THREADS)

clean-1:
	-rm -rf build-$(ENGINE)$(OPT_FLAG)-$(MODEL)-$(IO)-$(THREADS)

clean:
	-rm -rf build-* interpreter-* simplejit-* j2c-*

NO_JIT-%-JVM:
	echo "test $@ skipped"

NO_JIT-%-RealtimeJVM:
	echo "test $@ skipped"

NO_JIT-%-RealtimeJLThread:
	echo "test $@ skipped"

simplejit-simplejitopt-%: force
	$(SHOW) -n "testing $@ ..."
	$(MKTREE) build-$@
	-rm -f build-$@/*
	cd build-$@; $(GEN_OVM) -engine=$(ENGINE)	\
				$(OPT_FLAG)		\
				-model=$(MODEL)		\
				-threads=$(THREADS)	\
				-io=$(IO)		\
				-main=test.jvm.TestSuite   \
				-ud-reflective-methods='$(JVM_REFLECTION)' \
				$(REDIR)
	cd build-$@; ./ovm $(IMG) -doexectests -dotests	 \
		           test.jvm.TestSuite $(REDIRA)
	$(SHOW) passed



# Reflection usage within test and test.jvm packages.
JVM_REFLECTION=								   \
	Ltest/TestReflection$$MyClass; <init>:()V			   \
	Ltest/TestReflection$$MyClass; <init>:(Ljava/lang/String;)V	   \
	Ltest/TestReflection$$MyClass; <init>:(Ljava/lang/String;I)V	   \
	Ltest/TestReflection$$MyClass; methodA:(Ljava/lang/String;)I	   \
	Ltest/TestReflection$$MyClass; methodA:(Ljava/lang/String;I)I	   \
	Ltest/TestReflection$$MyClass; nonStaticAdd:(II)I		   \
	Gtest/TestReflection$$MyClass; staticAdd:(II)I			   \
	Ltest/jvm/TestFinalization$$Finalizable; <init>:()V		   \
	Ltest/jvm/TestReflection; <init>:(Ltest/jvm/TestBase$$Harness;ZZ)V \
	Ltest/jvm/TestReflection$$Helper; <init>:(ZZ)V 			   \
	Ltest/jvm/TestReflection; setField:(Z)V 			   \
	Gtest/jvm/TestReflection; staticMethod:(Ltest/jvm/TestReflection$$Helper;I)I

RTJVM_REFLECTION=							\
	$(JVM_REFLECTION)						\
	Ltest/rtjvm/TestMemoryAreas$$Finalizable; <init>:()V		\
	Ltest/rtjvm/TestMemoryAreas$$Cell; <init>:(II)V

# The space in the first line is critical to automated regression
# tests, it makes parsing way easier
%-JVM: force
	$(SHOW) -n "testing $@ ..."
	$(MKTREE) build-$@
	-rm -f build-$@/*
	cd build-$@; $(GEN_OVM) -engine=$(ENGINE)	\
				$(OPT_FLAG)		\
				-model=$(MODEL)		\
				-threads=$(THREADS)	\
				-transaction=$(TRANS)	\
				-io=$(IO)		\
				-main=test.jvm.TestSuite   \
				-ud-reflective-methods='$(JVM_REFLECTION)' \
				$(REDIR)
	cd build-$@; ./ovm $(IMG) -doexectests -dotests	 \
		           test.jvm.TestSuite $(REDIRA)
	$(SHOW) passed

%-RealtimeJVM: force
	$(SHOW) -n "testing $@ ..."
	$(MKTREE) build-$@
	-rm -f build-$@/*
	cd build-$@; $(GEN_OVM) -engine=$(ENGINE)	     \
				$(OPT_FLAG)		     \
				-model=$(MODEL)		     \
				-threads=$(THREADS)	     \
				-transaction=$(TRANS)	\
				-io=$(IO)		\
				-main=test.rtjvm.TestSuite   \
				-ud-reflective-methods='$(RTJVM_REFLECTION)' \
				$(REDIR)
	cd build-$@; ./ovm $(IMG) -doexectests -dotests	      \
			   test.rtjvm.TestSuite $(REDIRA)
	$(SHOW) passed

%-RealtimeJLThread: force
	$(SHOW) -n "testing $@ ..."
	$(MKTREE) build-$@
	-rm -f build-$@/*
	cd build-$@; $(GEN_OVM) -engine=$(ENGINE)	\
				$(OPT_FLAG)		\
				-model=$(MODEL)		\
				-io=$(IO)		\
				-threads=$(THREADS)	\
				-transaction=$(TRANS)	\
				$(REDIR)
	cd build-$@; ./ovm $(IMG) -doexectests -dotests $(REDIRA)
	cd build-$@; ./ovm $(IMG) -noforking -doexectests -dotests $(REDIRA)
	$(SHOW) passed


TEST=@-echo; time $(MAKE) -s REGRESS=true

# More thorough tests.
#
# There are various permutations of engine, threading, model and IO:
#  engine:    interpreter, simplejit or j2c
#  threading: JVM, real-time JVM, real-time JLThread (allows for multi-threaded
#                                                     ED tests)
#  model:  various combinations of allocators/GC, monitor support and object
#          layout
#  IO:  Various permutations of select() based, SIGIO based and polling based
#       I/O
# The last 2 tests run j2c on our largest configuration with optimization 
# turned up and again with optimization turned off.  
# 
# We try to avoid using precise garbage collectors with the interpreter
# because interpreter stack walking is an order of magnitude slower than
# it should be.
regression-test:
	$(TEST) ENGINE=interpreter		    \
		MODEL=MostlyCopyingSplitRegions-B_Mf_F_H \
		THREADS=RealtimeJVM		    \
		IO=SelectSockets_PollingOther
	$(TEST) ENGINE=interpreter		    \
		MODEL=MostlyCopyingSplitRegions-B_M_F_H \
		THREADS=RealtimeJVM		    \
		IO=SelectSockets_PollingOther
	$(TEST) ENGINE=interpreter		    \
		MODEL=MostlyCopyingSplitRegions-B_Mf_F_H \
		THREADS=RealtimeJVM		    \
		IO=SIGIOSockets_PollingOther
	$(TEST) ENGINE=interpreter		    \
		MODEL=MostlyCopyingSplitRegions-B_Mf_F_H \
		THREADS=RealtimeJVM		    \
		IO=SIGIOSockets_StallingFiles_PollingOther
	$(TEST) ENGINE=$(ARCH_JIT)		   \
		MODEL=MostlyCopyingSplitRegions-B_M_F_H \
		THREADS=RealtimeJVM		   \
		IO=SelectSockets_PollingOther
	$(TEST) ENGINE=j2c			   \
		MODEL=MostlyCopyingSplitRegions-B_Mf_F_H \
		THREADS=RealtimeJVM		   \
		TRANS=true			   \
		IO=SelectSockets_PollingOther
	$(TEST) ENGINE=j2c			   \
		MODEL=MostlyCopyingSplitRegions-B_M_F_H \
		THREADS=RealtimeJVM		   \
		TRANS=true			   \
		IO=SelectSockets_PollingOther
	$(TEST) ENGINE=j2c			   \
		MODEL=MostlyCopyingSplitRegions-B_Mf_F_H \
		THREADS=RealtimeJVM		   \
		TRANS=true			   \
		IO=SIGIOSockets_PollingOther
	$(TEST) ENGINE=j2c			   \
		MODEL=MostlyCopyingSplitRegions-B_Mf_F_H \
		THREADS=RealtimeJVM		   \
		TRANS=true			   \
		IO=SelectSockets_StallingFiles_PollingOther
	$(TEST) ENGINE=interpreter		\
		MODEL=SimpleSemiSpace-B_M_F_H	\
		THREADS=JVM		        \
		IO=SelectSockets_PollingOther
	$(TEST) ENGINE=$(ARCH_JIT)		\
		MODEL=MostlyCopyingWB-B_M_F_H	\
		THREADS=JVM		        \
		IO=SelectSockets_PollingOther
	$(TEST) ENGINE=$(ARCH_JIT)		\
		OPT_FLAG=-simplejitopt          \
		MODEL=MostlyCopyingWB-B_M_F_H	\
		THREADS=JVM		        \
		IO=SelectSockets_PollingOther
	$(TEST) ENGINE=j2c			\
		MODEL=JMTk-B_M_J_H		\
		THREADS=JVM		        \
		IO=SelectSockets_PollingOther
	$(TEST) ENGINE=interpreter		\
		MODEL=minimalMM-B_M		\
		THREADS=RealtimeJLThread	\
		IO=SelectSockets_PollingOther
	$(TEST) ENGINE=interpreter		\
		MODEL=minimalMM-B_M		\
		THREADS=RealtimeJLThread	\
		IO=SIGIOSockets_PollingOther
	$(TEST) ENGINE=$(ARCH_JIT)		\
		MODEL=MostlyCopyingSC-B_M_F_H	\
		THREADS=RealtimeJLThread	\
		IO=SelectSockets_PollingOther
	$(TEST) ENGINE=$(ARCH_JIT)		\
		MODEL=MostlyCopyingWB-B_Mf_F_H	\
		THREADS=JVM		        \
		IO=SelectSockets_PollingOther
	$(TEST) ENGINE=j2c			\
		MODEL=JMTk-B_M_J_H		\
		THREADS=RealtimeJLThread	\
		IO=SelectSockets_PollingOther
	$(TEST) ENGINE=j2c			\
		MODEL=JMTk-B_M_J_H		\
		THREADS=RealtimeJLThread	\
		IO=SIGIOSockets_PollingOther
	$(TEST) THREADS=JVM			\
		MODEL=MostlyCopyingWB-B_Mf_F_H 	\
	        OPT_FLAG=-opt=run		\
		IO=SelectSockets_PollingOther
	$(TEST) THREADS=RealtimeJVM		    \
		MODEL=MostlyCopyingSplitRegions-B_Mf_F_H \
		OPT_FLAG=-opt=run		    \
		IO=SelectSockets_PollingOther
	$(TEST) OPT_FLAG=-opt=none		    \
		MODEL=MostlyCopyingSplitRegions-B_M_F_H  \
		THREADS=RealtimeJVM		    \
		IO=SelectSockets_PollingOther

